(* Example for use of DPMI-Calls to serve a hardware interrupt
   from Delphi 1
   Change the follwing routines to adapt to special device:
     InitIRQ    (enable interrupt from device)
     ReleaseIRQ (disable interrupt from device)
     RMInt      (interrupt service routine)
   The following demo program shows the use for input from
   serial interface (COM-port with UART of type 8250). Only few
   changes are necessary to adapt to othjer devices (see below).

   J. Rathlev, Uni-Kiel, Apr. 1998

   see : W. Oney, Systemprogrammierung f. Windows
         DPMI-Spec. Vers. 0.9
*)
unit IntDemo;

interface

uses WinTypes, Messages, WinProcs, DPMI;

const
  errOK         = 0;    { no error on init }
  errIRQ        = 1;    { no interrupt specified }
  errSegment    = 2;    { lock segment error }

  wmCharReceived = wm_User+1; { message when character is received }

var
  IRQCnt   : word;  (* for testing *)
  LastChar : char;

function InitIRQ(APortAdr,AIntNr : word;
                 Handle : HWnd) : word;

procedure ReleaseIRQ;

implementation
const
  HwMsk   : byte = 0;    { IRQ Mask }
  IntCtlAdr = $20;       { Portaddress of Int.-Contr. }
  Int2CtlAdr = $A0;      { Portaddress of 2.Int.-Contr. }
  IntMskAdr = $21;       { Portaddress of Int.-Maskreg. }
  Int2MskAdr = $21;      { Portaddress of 2.Int.-Maskreg. }
  EOI = $20;             { end of interrupt }

  { 8250 Register }
  IER     =     1;       {  Interrupt Enable Register }
  IIR     =     2;       {  Interrupt Identification Register }
  LCR     =     3;       {  Line Control Register }
  MCR     =     4;       {  MODEM Control Register }
  LSR     =     5;       {  Line Status Register }
  MSR     =     6;       {  MODEM Status Register }

var
  PortAdr,                { I/O-Port }
  HwInt       : word;     { IRQ Number }
  I2Ctrl      : boolean;  { 1./2. IRQ-Controller }
  OldInt      : pointer;  { old real mode interrupt routine }
  RmRegs      : TRealModeRegs;  { real mode register structure }
  MainHandle  : HWnd;      { handle of main program }

(* -------------------------------------------------------------------------
   Real Mode Interrupt Routine *)
procedure RMInt (SI,DI,DS,ES,BP : word); interrupt;
var
  Status : byte;
  Stack : ^StackArr;
  rmr   : ^TRealModeRegs;
begin
  inc (IRQCnt);   (* count for demo *)
  (* device interrupt service
     Printer port:
       Port[$378+1]:=$11;       { set bit 1 (STB) }
       Port[$378+1]:=$10;       { clear bit 1 (STB) -> clear int. }
     COM-Port:
       St:=Port[$3F8/2F8+LSR] and $1E;  { line status }
       Ch:=chr(Port[$3F8/2F8]);         { received character }
     *)
  Status:=Port[PortAdr+LSR] and $1E;
  LastChar:=chr(Port[PortAdr]);       { get received character }
  Port[PortAdr]:=ord(LastChar);       { echo }

  (* Post message for main program *)
  PostMessage (MainHandle,wmCharReceived,ord(LastChar),IRQCnt);

  (* device independent part of interrupt routine *)
  Stack:=Ptr(ds,si);
  rmr:=Ptr(es,di);
  Port[IntCtlAdr]:=EOI;
  if I2Ctrl then Port[Int2CtlAdr]:=EOI;

  (* set real mode cp:ip, update stack pointer
     (DPMI-Spez. Vers. 0.9, chap. 11.4 *)
  with rmr^ do begin
    ip:=Stack^[0];
    cs:=Stack^[1];
    flags:=Stack^[2];
    inc(sp,6);
    end;
  end;

(* -------------------------------------------------------------------------
  Init interface for use of hardware interrupt:
  APortAdr:  address of port from which interrupt is generated
             (e.g. $3F8 (COM1),
                   $2F8 (COM2),
                   $378 (LPT1),
                   $278 (LPT2)
  AIntNr:    number of used hardware interrupt
             (e.g. 4 (COM1),
                   3 (COM2),
                   7 (LPT1),
                   5 (LPT2))
             interrupt from 2nd controller (9,10,11,..) are supported
  Handle:    handle of main program *)
function InitIRQ(APortAdr,AIntNr : word;
                 Handle : HWnd) : word;
var
  i  : integer ;
  cb : byte;
begin
  if (LockSegment(CSeg)<>0) and (LockSegment(DSeg)<>0) then begin
    if AIntNr>0 then begin
      (* Enable interrupt service for device:
         Printer port:
           Port[$378+2]:=$10;       { set bit 4 (IEN) }
         COM-Port (COM1/COM2):
           Port[$3F8/2F8+MCR]:=$0B;   { set DTR, RTS and OUT2 }
           Port[$3F8/2F8+IER]:=1;     { enable interrupt for received data }
         *)

      PortAdr:=APortAdr;
      (* set baud rate etc. *)
      Port[PortAdr+LCR]:=$80;       { set DLAB }
      Port[PortAdr]:=24;            { 4800 baud }
      Port[PortAdr+1]:=0;
      Port[PortAdr+LCR]:=3;         { n,8,1 }

      cb:=Port[PortAdr];            { clear receiver buffer reg. }

      (* set DTR, RTS and OUT2, enable receive interrupt *)
      Port[PortAdr+MCR]:=$0B;
      Port[PortAdr+IER]:=1;
      LastChar:=#0;
      IRQCnt:=0;

      (* save interrupt number and mask *)
      HwInt:=AIntNr; I2Ctrl:=HwInt>=8;
      if I2Ctrl then begin   (* 2. Contr. *)
       dec(AIntNr,8);
       InitRmInt (HwInt+$68,@RMInt,RmRegs,OldInt)
       end
      else InitRmInt (HwInt+8,@RMInt,RmRegs,OldInt);
      { set interrupt mask }
      HwMsk:=1 shl AIntNr;
      if I2Ctrl then Port[Int2MskAdr]:=Port[Int2MskAdr] and (not HwMsk)
      else Port[IntMskAdr]:=Port[IntMskAdr] and (not HwMsk);
      SetInt;
      InitIRQ:=errOK;
      MainHandle:=Handle;
      end
    else InitIRQ:=errIRQ;
    end
  else InitIRQ:=errSegment;
  end;

(* -------------------------------------------------------------------------
   Release hardware interrupt *)
procedure ReleaseIRQ;
begin
  if HwMsk>0 then begin (* restore interrupt vector *)
    (* Disable interrupt service for device:
       Printer port:
         Port[$378+2]:=0;         { clear bit 4 (IEN) }
       COM-Port:
         Port[$3F8/2F8+1]:=0;     { disable interrupt }
         Port[$3F8/2F8+4]:=0;     { clear DTR, RTS and OUT2 }
       *)
    Port[PortAdr+IER]:=0;
    Port[PortAdr+MCR]:=0;
    if I2Ctrl then begin
      Port[Int2MskAdr]:=Port[Int2MskAdr] or HwMsk;
      RlsRmInt (HwInt+$68,OldInt);
      end
    else begin
      Port[IntMskAdr]:=Port[IntMskAdr] or HwMsk;
      RlsRmInt (HwInt+8,OldInt);
      end;
    HwMsk:=0;
    end;
  UnLockSegment(CSeg); UnLockSegment(DSeg);
  end;

end.
